{ created 5/99 by Brian Gillingham for a nice enhanced GUI control, the
  TEXTUREPANEL!

  Should act as a standard panel in all respects except for the color that
  is filled in the Paint procedure is a FillRect procedure using a textured
  Brush.Bitmap of the Panel's color.

  properties:
    NoiseAmount  (0..100)                     - the amount of random color
    NoiseMonoChrome (True / False)            - is random color all one Hue ?
    TextureSize  (4..240)                     - the size of the Brush.Bitmap
    VerticalAlign (vaTop, vaCenter, vaBottom) - for Vertical control over where
    	the caption is printed

}

unit TexturePanel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

const
 	DarkStrength = 0.82;
  StrongBlend = 52;
  WeakBlend = 36;

type
  TVerticalAlign = (vaTop, vaCenter, vaBottom);

  TTexturePanel = class(TPanel)
  private
    { Private declarations }
    FNoiseAmount     : integer;
    FNoiseMonoChrome : boolean;
    FTextureSize     : integer;
		FVerticalAlign   : TVerticalAlign;
    procedure SetNoiseAmount(Value: integer);
    procedure SetNoiseMonoChrome(Value: boolean);
    procedure SetTextureSize(Value: integer);
    procedure SetVerticalAlign(Value: TVerticalAlign);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;

    { Field variable Declarations }
  protected
    { Protected declarations }
    tmpBitmap          : TBitmap;
    function MakeBrushBitmapSize(Colr: longInt; BrushSize: integer): TBitmap;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Paint; override;
    procedure RefreshBrush;
    function MakeBrush(Colr: longInt): TBitmap;
  published
    { Published declarations }
    property NoiseAmount: integer read FNoiseAmount write SetNoiseAmount
    	default 48;
    property NoiseMonoChrome: boolean read FNoiseMonoChrome write
    	SetNoiseMonoChrome;
    property TextureSize: integer read FTextureSize write SetTextureSize
    	default 40;
    property VerticalAlign: TVerticalAlign read FVerticalAlign write
    	SetVerticalAlign default vaCenter;
  end;

function BlendColors(const Color1, Color2: longInt; Opacity: integer): longInt;
function Highlight(Colr: longInt): longInt;
function MyRandom(Number: integer): integer;
function NoiseColor(const Color1: longInt; Amount: integer;
	MonoChrome: boolean): longInt;
function Shadow(Colr: longInt): longInt;
procedure Register;
procedure DoDrawText(Canvas: TCanvas; var Rect: TRect; Flags: integer;
	text: string; ShowAccelChar: boolean);

const
  VerAlign: array[TVerticalAlign] of Word	= (0, DT_VCENTER or DT_SINGLELINE,
  	DT_BOTTOM or DT_SINGLELINE);

implementation

procedure Register;
begin
  RegisterComponents('Mine', [TTexturePanel]);
end; { Register }
{------------------------------------------------------------------------------}

procedure DoDrawText(Canvas: TCanvas; var Rect: TRect; Flags: integer;
	text: string; ShowAccelChar: boolean);
begin
  if (Flags and DT_CALCRECT <> 0) and ((Text = '') or ShowAccelChar and
    (Text[1] = '&') and (Text[2] = #0)) then Text := Text + ' ';
  if not ShowAccelChar then Flags := Flags or DT_NOPREFIX;
  DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, Flags);
end; { DoDrawText }
{------------------------------------------------------------------------------}

function BlendColors(const Color1, Color2: longInt; Opacity: integer): longInt;
var
	R, R1, R2, G, G1, G2, B, B1, B2: integer;
begin
	Opacity:= abs(Opacity);
	if Opacity > 100 then
  	Opacity:= 100;
	R1:= GetRValue(ColorToRGB(Color1));
	G1:= GetGValue(ColorToRGB(Color1));
	B1:= GetBValue(ColorToRGB(Color1));
	R2:= GetRValue(ColorToRGB(Color2));
	G2:= GetGValue(ColorToRGB(Color2));
	B2:= GetBValue(ColorToRGB(Color2));
  R:= trunc(R1 * Opacity / 100) + trunc(R2 * (100 - Opacity) / 100);
  G:= trunc(G1 * Opacity / 100) + trunc(G2 * (100 - Opacity) / 100);
  B:= trunc(B1 * Opacity / 100) + trunc(B2 * (100 - Opacity) / 100);
  Result := RGB(R, G, B);
end; { BlendColors }
{------------------------------------------------------------------------------}

function Highlight(Colr: longInt): longInt;
var
	avg, r, g, b: integer;
  tmp: longInt;
begin
  r:= GetRValue(Colr);
  g:= GetGValue(Colr);
  b:= GetBValue(Colr);
  avg:= (r + g + b) div 3;
	r:= (255 + 255 + avg + r) div 4;
  g:= (255 + 255 + avg + g) div 4;
  b:= (255 + 255 + avg + b) div 4;
  tmp:= RGB(r, g, b);
	Result:= BlendColors(Colr, tmp, WeakBlend);
end; { Highlight }
{------------------------------------------------------------------------------}

function MyRandom(Number: integer): integer;
begin
	result:= Random(Number);
    if random(2)<1 then
	    result:= abs(Number - result);
end; { MyRandom }
{------------------------------------------------------------------------------}

function NoiseColor(const Color1: longInt; Amount: integer;
	MonoChrome: boolean): longInt;
var
	R, R1, G, G1, B, B1: integer;
  Half: integer;
  pctR, pctG, pctB: real;
begin
	R1:= GetRValue(ColorToRGB(Color1));
	G1:= GetGValue(ColorToRGB(Color1));
	B1:= GetBValue(ColorToRGB(Color1));

  if MonoChrome then
  begin
    pctR:= R1 / 256;
    pctG:= G1 / 256;
    pctB:= B1 / 256;

    Amount:= abs(Amount);
    if Amount > 100 then
      Amount:= 100;
    Half:= Amount div 2;

    R:= abs(R1 + Random(trunc(pctR * Amount) - trunc(pctR * Half)));
    G:= abs(G1 + Random(trunc(pctG * Amount) - trunc(pctG * Half)));
    B:= abs(B1 + Random(trunc(pctB * Amount) - trunc(pctB * Half)));

    if R>255 then R:= 255;
    if G>255 then G:= 255;
    if B>255 then B:= 255;

		Result:= RGB(R, G, B);
  end
  else begin
    Amount:= abs(Amount);
    if Amount > 100 then
      Amount:= 100;
    Half:= Amount div 2;
    R:= abs(R1 + Random(Amount) - Half);
    G:= abs(G1 + Random(Amount) - Half);
    B:= abs(B1 + Random(Amount) - Half);
    if R>255 then R:= 255;
    if G>255 then G:= 255;
    if B>255 then B:= 255;

    Result := RGB(R, G, B);
  end;
end; { NoiseColor }
{------------------------------------------------------------------------------}

function Shadow(Colr: longInt): longInt;
var
	r, g, b: integer;
  tmp: longInt;
begin
  r:= GetRValue(Colr);
  g:= GetGValue(Colr);
  b:= GetBValue(Colr);
	tmp:= RGB(trunc(DarkStrength*r), trunc(DarkStrength*g),
  	trunc(DarkStrength*b));
	Result:= BlendColors(Colr, tmp, StrongBlend);
end; { Shadow }
{------------------------------------------------------------------------------}

{ TTexturePanel }

constructor TTexturePanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  { set the default values }
  FNoiseAmount:= 48;
  FTextureSize:= 40;
  FVerticalAlign:= vaCenter;
  randomize;

  ControlStyle := ControlStyle + [ csOpaque, csAcceptsControls, csCaptureMouse,
  	csClickEvents];
  RefreshBrush;
end;

destructor TTexturePanel.Destroy;
begin
  { free all objects }
  inherited Destroy;
end;

function TTexturePanel.MakeBrush(Colr: longInt): TBitmap;
var
  C: longInt;
begin
	C:= ColorToRGB(Colr);
  result:= MakeBrushBitmapSize(C, FTextureSize);
end;

procedure TTexturePanel.Paint;
const
  Alignments: array[TAlignment] of Longint = (DT_LEFT, DT_RIGHT, DT_CENTER);
var
  LocalRect: TRect;
  TopColor, BottomColor: TColor;
  Flags: Longint;
  Indent, BevelWidthMultiplier: integer;
  AdjRect: TRect;

  procedure AdjustColors(Bevel: TPanelBevel);
  begin
    TopColor := Highlight(Color);
    if Bevel = bvLowered then TopColor := Shadow(Color);
    BottomColor := Shadow(Color);
    if Bevel = bvLowered then BottomColor := Highlight(Color);
  end;

begin
	Canvas.Lock;
  try
    LocalRect := GetClientRect;
    if (BevelOuter = bvNone) and (BevelInner = bvNone) then
      BevelWidthMultiplier:= 0
    else
      if (BevelOuter = bvNone) or (BevelInner = bvNone) then
        BevelWidthMultiplier:= 1
      else
        BevelWidthMultiplier:= 2;
    if BorderStyle = bsSingle then
    begin
      Indent:= 1 + BevelWidthMultiplier*BevelWidth;
      Frame3D(Canvas, LocalRect, clBlack, clWhite, 1);
    end
    else
      Indent:= BevelWidthMultiplier*BevelWidth;
    { init some variables }
    AdjRect:= Rect( 0, 0, (LocalRect.Right - LocalRect.Left), (LocalRect.Bottom -
      LocalRect.Top) );

    if BevelOuter <> bvNone then
    begin
      AdjustColors(BevelOuter);
      Frame3D(Canvas, LocalRect, TopColor, BottomColor, BevelWidth);
    end;
    Frame3D(Canvas, LocalRect, Color, Color, BorderWidth);
    if BevelInner <> bvNone then
    begin
      AdjustColors(BevelInner);
      Frame3D(Canvas, LocalRect, TopColor, BottomColor, BevelWidth);
    end;
    with Canvas do
    begin
      Brush.Bitmap:= tmpBitmap;
      FillRect(Rect(Indent,Indent,Width-Indent,Height-Indent));

      Brush.Style := bsClear;
      Font := Self.Font;
      Flags := DT_EXPANDTABS or Alignments[Alignment] or
      	VerAlign[FVerticalAlign];
      Flags := DrawTextBiDiModeFlags(Flags);

      DoDrawText(Canvas, AdjRect, Flags, Caption, false);
    end; { with Canvas do }
  finally
  	Canvas.Unlock;
  end;
end;

procedure TTexturePanel.CMColorChanged(var Message: TMessage);
begin
  inherited;
  RefreshBrush;
end;

procedure TTexturePanel.SetVerticalAlign(Value: TVerticalAlign);
begin
	if Value<>FVerticalAlign then
  begin
		FVerticalAlign:= Value;
    invalidate;
  end;
end;

procedure TTexturePanel.SetTextureSize(Value: integer);
begin
	if Value<>FTextureSize then
  begin
  	if Value < 4 then
    begin
    	ShowMessage('Minimum Texture Size of' + #13 +
      	'  4 has been substituted');
      SetTextureSize(4);
      exit;
    end;

  	if Value > 240 then
    begin
    	ShowMessage('Maximum Texture Size of' + #13 +
      	'  240 has been substituted');
      SetTextureSize(240);
      exit;
    end;
		FTextureSize:= Value;
    RefreshBrush;
  end;
end;

procedure TTexturePanel.SetNoiseAmount(Value: integer);
begin
	if Value<>FNoiseAmount then
  begin
  	if Value > 100 then
    begin
    	ShowMessage('Maximum Noise Amount of' + #13 +
      	'  100 has been substituted');
      SetNoiseAmount(100);
      exit;
    end;
		FNoiseAmount:= abs(Value);
    RefreshBrush;
  end;
end;

procedure TTexturePanel.SetNoiseMonoChrome(Value: boolean);
begin
	if Value<>FNoiseMonoChrome then
  begin
		FNoiseMonoChrome:= Value;
    RefreshBrush;
  end;
end;

procedure TTexturePanel.RefreshBrush;
begin
	tmpBitmap:= MakeBrush(Color);
  invalidate;
end;

function TTexturePanel.MakeBrushBitmapSize(Colr: longInt;
	BrushSize: integer): TBitmap;
var
	BMP: TBitmap;
	NumDots, dots, x, y: integer;
  ColrShadow, ColrHighlight, TempC: longInt;
begin
	Screen.Cursor:= crHourglass;
	NumDots:= BrushSize * (NoiseAmount+BrushSize div 4);
	Colr:= ColorToRGB(Colr);
  ColrShadow:= Shadow(Colr);
  ColrHighlight:= Highlight(Colr);

	randomize;
	BMP:= TBitmap.create;
  try
    BMP.Width:= BrushSize;
    BMP.Height:= BrushSize;
    BMP.Canvas.Brush.Color:= Colr;
    BMP.Canvas.FillRect(Rect(0,0,BrushSize,BrushSize));
    dots:= 0;
    BMP.Canvas.Pen.Width:= 1;
    x:= 0;
    repeat
    	y:= x mod 2;
    	repeat
	      BMP.Canvas.Pixels[x,y]:= NoiseColor(Colr, FNoiseAmount, FNoiseMonoChrome);
        inc(y, random(2) + 2);
      until y>BrushSize;
    	inc(x, 2);
    until X>BrushSize;
    repeat
      {one random MEDIUM dot}
      x:= MyRandom(BrushSize);
      y:= MyRandom(BrushSize);
      TempC:= NoiseColor(Colr, FNoiseAmount div 4, FNoiseMonoChrome);
      BMP.Canvas.Pixels[x,y]:= TempC;
      BMP.Canvas.Pixels[x+1,y]:= BlendColors(TempC, ColrShadow, WeakBlend);
      BMP.Canvas.Pixels[x-1,y]:= BlendColors(TempC, ColrHighlight, WeakBlend);
      BMP.Canvas.Pixels[x,y+1]:= BlendColors(TempC, ColrShadow, WeakBlend);
      BMP.Canvas.Pixels[x,y-1]:= BlendColors(TempC, ColrHighlight, WeakBlend);

      {one random LARGE dot}
      x:= MyRandom(BrushSize);
      y:= MyRandom(BrushSize);
      TempC:= NoiseColor(Colr, FNoiseAmount div 4, FNoiseMonoChrome);
      BMP.Canvas.Pixels[x,y]:= TempC;
      BMP.Canvas.Pixels[x+1,y]:= BlendColors(TempC, ColrShadow, WeakBlend);
      BMP.Canvas.Pixels[x-1,y]:= BlendColors(TempC, ColrHighlight, WeakBlend);
      BMP.Canvas.Pixels[x,y+1]:= BlendColors(TempC, ColrShadow, WeakBlend);
      BMP.Canvas.Pixels[x,y-1]:= BlendColors(TempC, ColrHighlight, WeakBlend);
      BMP.Canvas.Pixels[x-1,y-1]:= BlendColors(TempC, ColrHighlight, StrongBlend);
      BMP.Canvas.Pixels[x-1,y+1]:= BlendColors(TempC, Colr, WeakBlend);
      BMP.Canvas.Pixels[x+1,y+1]:= BlendColors(TempC, ColrShadow, StrongBlend);
      BMP.Canvas.Pixels[x+1,y-1]:= BlendColors(TempC, Colr, WeakBlend);

      {another random LARGE dot}
      x:= MyRandom(BrushSize);
      y:= MyRandom(BrushSize);
      TempC:= NoiseColor(Colr, FNoiseAmount div 4, FNoiseMonoChrome);
      BMP.Canvas.Pixels[x,y]:= TempC;
      BMP.Canvas.Pixels[x+1,y]:= BlendColors(TempC, ColrShadow, WeakBlend);
      BMP.Canvas.Pixels[x-1,y]:= BlendColors(TempC, ColrHighlight, WeakBlend);
      BMP.Canvas.Pixels[x,y+1]:= BlendColors(TempC, ColrShadow, WeakBlend);
      BMP.Canvas.Pixels[x,y-1]:= BlendColors(TempC, ColrHighlight, WeakBlend);
      BMP.Canvas.Pixels[x-1,y-1]:= BlendColors(TempC, ColrHighlight, StrongBlend);
      BMP.Canvas.Pixels[x-1,y+1]:= BlendColors(TempC, Colr, WeakBlend);
      BMP.Canvas.Pixels[x+1,y+1]:= BlendColors(TempC, ColrShadow, StrongBlend);
      BMP.Canvas.Pixels[x+1,y-1]:= BlendColors(TempC, Colr, WeakBlend);

      {one random MEDIUM dot}
      x:= MyRandom(BrushSize);
      y:= MyRandom(BrushSize);
      TempC:= NoiseColor(Colr, FNoiseAmount div 4, FNoiseMonoChrome);
      BMP.Canvas.Pixels[x,y]:= TempC;
      BMP.Canvas.Pixels[x+1,y]:= BlendColors(TempC, ColrShadow, WeakBlend);
      BMP.Canvas.Pixels[x-1,y]:= BlendColors(TempC, ColrHighlight, WeakBlend);
      BMP.Canvas.Pixels[x,y+1]:= BlendColors(TempC, ColrShadow, WeakBlend);
      BMP.Canvas.Pixels[x,y-1]:= BlendColors(TempC, ColrHighlight, WeakBlend);
      {one random SMALL dot}
      x:= MyRandom(BrushSize);
      y:= MyRandom(BrushSize);
      TempC:= NoiseColor(Colr, FNoiseAmount div 4, FNoiseMonoChrome);
      BMP.Canvas.Pixels[x,y]:= TempC;
      inc(dots);
    until dots>(NumDots) div 80;

    Result:= Bmp;
  finally
		Screen.Cursor:= crDefault;
  end;
end; { MakeBrushBitmapSize }
{------------------------------------------------------------------------------}

end.
